# -*- coding: utf-8 -*-
"""
"""

import sys
from . import __version__
from .log import logger


class ExitCode():
    EXPECTED_EXIT = 0
    
    NO_CONFIG_FOUND = 1
    INVALID_CONFIG = 2
    DUPLICATE_CONFIG_SECTIONS = 3
    CONFIG_WRITE_ERROR = 4
    NO_CURRENT_VERSION = 5
    
    GIT_UNUSABLE = 10
    GIT_ROOT_ERROR = 11
    GIT_DESCRIBE_ERROR = 12
    
    VERSION_PARSE_ERROR = 20
    VERSION_METADATA_ERROR = 21
    VERSION_BUMP_ERROR = 22
    VERSION_PEP440_ERROR = 23
    VERSION_GIT_ERROR = 24

    SOURCE_CODE_FILE_OPEN_ERROR = 30
    SOURCE_CODE_FILE_MISSING = 31
    USER_SUPPLIED_REGEX_ERROR = 32


class IcanException(Exception):
    def __init__(self, *args, **kwargs):
        """ """
        self.output_method = logger.critical
        self.exit_code = self.__class__.exit_code
        self.msg = ""
        self.e = []
        
        # Set the message, typically at init
        if args:
            if isinstance(args[0], Exception):
                self.msg = str(args[0])

                _e = args[0]
                if hasattr(_e, 'stderr'):
                    stderr = getattr(_e, 'stderr')
                    stderr = stderr.decode('utf-8')
                    stderr = stderr.rstrip('\n')
                    self.e.append(stderr)
                if hasattr(_e, 'returncode'):
                    returncode = getattr(_e, 'returncode')
                    self.e.append(returncode)
            else:
                self.msg = args[0]
        elif hasattr(self.__class__, "msg"):
            self.msg = self.__class__.msg

        # Merge the exit_code, msg, and version
        m = f'{self.msg} (code-{self.exit_code}) v{__version__}'
        self.msg = m

    def __str__(self):
        return self.msg

class DryRunExit(IcanException):
    pass

########################

class NoConfigFound(IcanException):
    exit_code = ExitCode.NO_CONFIG_FOUND
    msg = 'Cannot find ican config.  Maybe try `ican init`'

class InvalidConfig(IcanException):
    exit_code = ExitCode.INVALID_CONFIG

class DuplicateConfigSections(IcanException):
    exit_code = ExitCode.DUPLICATE_CONFIG_SECTIONS
    msg = 'Duplicate sections in config file. Did you create file/pipeline labels?'

class ConfigWriteError(IcanException):
    exit_code = ExitCode.CONFIG_WRITE_ERROR

class NoCurrentVersion(IcanException):
    exit_code = ExitCode.NO_CURRENT_VERSION
    msg = (
        "[NO_VERSION_SPECIFIED]\n"
        "Check if current version is specified in config file, like:\n"
        "version = 0.4.3\n"
    )

########################

class GitUnusable(IcanException):
    exit_code = ExitCode.GIT_UNUSABLE

class GitRootError(IcanException):
    exit_code = ExitCode.GIT_ROOT_ERROR

class GitDescribeError(IcanException):
    exit_code = ExitCode.GIT_DESCRIBE_ERROR

########################

class VersionParseError(IcanException):
    exit_code = ExitCode.VERSION_PARSE_ERROR

class VersionMetadataError(IcanException):
    exit_code = ExitCode.VERSION_METADATA_ERROR

class VersionBumpError(IcanException):
    exit_code = ExitCode.VERSION_BUMP_ERROR

class VersionPep440Error(IcanException):
    exit_code = ExitCode.VERSION_PEP440_ERROR

class VersionGitError(IcanException):
    exit_code = ExitCode.VERSION_GIT_ERROR

########################

class SourceCodeFileOpenError(IcanException):
    exit_code = ExitCode.SOURCE_CODE_FILE_OPEN_ERROR

class SourceCodeFileMissing(IcanException):
    exit_code = ExitCode.SOURCE_CODE_FILE_MISSING

class UserSuppliedRegexError(IcanException):
    exit_code = ExitCode.USER_SUPPLIED_REGEX_ERROR

########################



# -*- coding: utf-8 -*-
"""
"""
import re
from pathlib import Path

from .log import logger
from .log import ok_to_write
from . import exceptions
from .exceptions import UserSuppliedRegexError
from .exceptions import SourceCodeFileOpenError
from .exceptions import SourceCodeFileMissing

#######################################
#
#   SourceCode - represents a
#     file that we are updating.
#
#######################################


class SourceCode(object):

    VARIABLE_RE = r"{{var}}\s*=\s*(?P<quote>[\'\"])(?P<version>.+)(?P=quote)"

    def __init__(self, label, file, style='semantic', variable=None, regex=None):
        self.label = f'{label.upper()}[{file}]'
        self.file = Path(file)
        self.variable = variable
        self.style = style
        self.regex = regex

        self.updated = False
        self.valid = False
        
        if not self.file.exists():
            raise SourceCodeFileMissing(
                f'config references non existant file: {self.file}'
            )
        self.valid = True

        if variable is not None:
            self.regex = SourceCode.VARIABLE_RE.replace("{{var}}", variable)
            logger.debug(f'regex generated for {variable}')

        if self.regex:
            try:
                self.compiled = re.compile(self.regex)
            except Exception as e:
                msg = f'Error compiling regex: {self.regex}'
                raise UserSuppliedRegexError(msg)

    def _to_raw_string(self, str):
        return fr"{string}"

    def _replacement(self, match):
        line = match.group(0)
        old_version = match.group('version')
        new_line = line.replace(old_version, self.new_version)
        
        return new_line

    def update(self, version):
        """
        his method performs an inplace file update.  
        Args:
            filename: The file to run the substitution on
        Returns:
            True if all is successful.  Filename will be updated
            with new version if found.
        """

        self.new_version = getattr(version, self.style)
        logger.debug(
            f'{self.label} - updating to {self.new_version}'
        )

        try:
            f = self.file.open('r+')
        except OSError:
            raise SourceCodeFileOpenError(f'Error opening {self.file}')

        with self.file.open('r+') as f:
            # Read entire file into string
            original = f.read()

            # Regex search
            updated, n = self.compiled.subn(self._replacement, original, count=1)

            # Check if we found a match or not
            if n == 0:
                logger.debug(f'{self.label} - NO MATCHES!')
                return
            logger.debug(f'{self.label} - found {n} matches')

            # Write the updated file
            if ok_to_write():
                f.seek(0)
                f.write(updated)
                f.truncate()

        self.updated = True
        logger.debug(f'{self.label} - update COMPLETE')
        return True



# -*- coding: utf-8 -*-
"""
from tempfile import NamedTemporaryFile
"""

import argparse
import sys

from . import __version__
from .ican import Ican
from .log import logger
from .log import setup_console_handler
from .emojis import rnd_good_emoji
from .exceptions import IcanException


#===============================
#
#  CLI Class
#
#===============================


class CLI(object):
    usage='''ican <command> [<args>]

Some of our most popular commands:
   bump [PART]      increment the PART of the version
                    [minor, major, patch, prerelease, build]
   show [STYLE]     display current version with STYLE
                    [semantic, public, pep440, git]
   init             initialize the current directory with a 
                    config file
'''

    def __init__(self):
        self._register_excepthook()
        self.i = Ican()
        self.aliases = self.i.config.aliases

        parser = argparse.ArgumentParser(
            description='ican - version bumper and lightweight build pipelines',
            usage=CLI.usage,
            prog='ican'
        )
        parser.add_argument('command', help='command-specific-arguments')    # need
        parser.add_argument(
            '--version', 
            action='version',
            help='display ican version',
            version=f'ican v{__version__}'
        )

        # if config.alias was used, insert it into sys.argv
        command = sys.argv[1]
        if self.aliases.get(command):
            built_in = self.aliases.get(command)
            sys.argv.pop(1)   #delete the alias command
            sys.argv[1:1] = built_in

        # now parse our upsated args
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            # no method for the command
            logger.error('Unrecognized command')
            parser.print_help()
            exit(1)

        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()
        
        return


    def _register_excepthook(self):
        self._original_excepthook = sys.excepthook
        sys.excepthook = self._excepthook

    def _excepthook(self, type, value, tracekback, debug=False):
        """ """
        if isinstance(value, IcanException):
            if value.msg:
                value.output_method(value.msg)
            if value.e:
                for line in value.e:
                    value.output_method(line)
            if debug:
                self._original_excepthook(type, value, tracekback)
            exit_code = value.exit_code
            sys.exit(exit_code)
        else:
            self._original_excepthook(type, value, tracekback)

    def fetch_args(self, parser):
        """
        now that we're inside a subcommand, ignore the first
        TWO argvs, ie the command (git) and the subcommand (commit)
        """
        
        parser.add_argument(
            '--dry-run',
            help='files will not be written - best with --verbose',
            action="store_true"
        )
        parser.add_argument(
            '--verbose',
            help='display all debug information available',
            action="store_true"
        )
        args = vars(parser.parse_args(sys.argv[2:]))
        setup_console_handler(args['verbose'], args['dry_run'])

        return args

    def bump(self):
        parser = argparse.ArgumentParser(
            description='increment the [PART] of the version')
        parser.add_argument(
            "part", 
            nargs='?',
            default='build',
            choices=['major', 'minor', 'patch', 'prerelease', 'build'],
            help="what to bump"
        )
        args = self.fetch_args(parser)
        part = args['part']

        self.i.bump(part.lower())
        logger.debug('bump() COMPLETE')
        logger.warning(f'Version: {self.i.version.semantic}')

        return

    def show(self):
        parser = argparse.ArgumentParser(
            description='show the [STYLE] of current version')
        parser.add_argument(
            "style", 
            nargs='?',
            default='semantic',
            choices=['semantic', 'public', 'pep440', 'git'],
            help="version style to show"
        )
        args = self.fetch_args(parser)

        v = self.i.show(args['style'])
        logger.debug('show() COMPLETE')
        logger.warning(v)

        return

    def init(self):
        parser = argparse.ArgumentParser(
            description='initialize your project in the current directory')
        args = self.fetch_args(parser)

        del self.i
        self.i = Ican(init=True)
        logger.warning('init COMPLETE')

        return

    def test(self):
        parser = argparse.ArgumentParser(description='test')
        parser.add_argument(
            "first",
            nargs='?',
            help="first test arg"
        )
        args = self.fetch_args(parser)
        print(f'10-4 with arg {args["first"]}')


def entry():
    CLI()




# -*- coding: utf-8 -*-

import os
from pathlib import Path

from .config import Config
from .version import Version
from .git import Git
from .log import logger
from .log import ok_to_write
from .emojis import rnd_good_emoji
from .exceptions import GitDescribeError

#######################################
#
#   Bump Class
#
#######################################


class Ican(object):
    """
    Object which will orchestrate entire program
    """

    def __init__(self, init=False):
        self.version = None
        self.git = None
        self.config = None

        logger.debug('Verbose output selected.')

        # Git init - Do this early incase we need git.root
        self.git = Git()

        # Create config obj.  If init, set defaults.
        # Otherwise, search for existing config file.
        self.config = Config(self.git.root)
        if init:
            self.config.init()
        else:
            self.config.search_for_config()

        # Now we have default or existing config, we can parse
        self.config.parse()

        # Now config is parsed.  We can parse from config
        self.version = Version.parse(self.config.current_version)
        logger.debug(f'discovered {self.version.semantic} @ CONFIG.version')

        try:
            self.version._git_metadata = self.git.describe()
        except GitDescribeError as e:
            logger.info(e)
            logger.info('Git style versions will be disabled.')
            logger.info('Possibly this is a new repo with no tags.')
            self.git.disable()

        else:
            logger.debug(f'discovered {self.version.git} @ GIT.version')

        return

    def show(self, style):
        """
        Show the <STYLE> version
        """

        v = getattr(self.version, style)
        if v is None:
            return f'version STYLE: {style} not available'
        return v

    def bump(self, part):
        """
        This is pretty much the full process
        """

        logger.debug(f'beginning bump of <{part.upper()}>')

        # Use the Version API to bump 'part'
        self.version.bump(part)
        logger.debug(
            f'new value of <{part.upper()}> - {getattr(self.version, part)}'
        )

        # Update the user's files with new version
        for file in self.config.source_files:
            file.update(self.version)

        # Pipeline
        if self.version.new_release:
            if self.config.pipelines.get('release'):
                self.run_pipeline('release')
        elif part == 'build':
            if self.config.pipelines.get('build'):
                self.run_pipeline('build')

        # Once all else is successful, persist the new version
        self.config.persist_version(self.version.semantic)

        return self

    def run_pipeline(self, label):
        logger.debug('RELEASE pipeline triggered')

        pl = self.config.pipelines.get(label)
        ctx = {}
        ctx['tag'] = self.version.tag
        ctx['msg'] = f'auto commit for {self.version.tag}.'
        pl.run(ctx)



import logging
import sys
from . import __version__
from .emojis import rnd_good_emoji
from .emojis import rnd_bad_emoji


__all__ = [
    'logger',
    'DEBUG',
    'INFO',
    'WARNING',
    'ERROR'
]


class CustomFormatter(logging.Formatter):
    """Logging colored formatter, adapted from https://stackoverflow.com/a/56944256/3638629"""
    
    BLACK = '\u001b[30;1m'
    RED = '\u001b[31;1m'
    GREEN = '\u001b[32;1m'
    YELLOW = '\u001b[33;1m'
    BLUE = '\u001b[34;7m'
    MAGENTA = '\u001b[35;1m'
    CYAN = '\u001b[36;1m'
    WHITE = '\u001b[37;1m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    BLINK = '\033[5m'
    INVERT = '\033[7m'
    RESET = '\u001b[0m'

    def __init__(self, fmt, style='%', emoji=True):
        super().__init__()
        self.fmt = fmt
        self.plain_fmt = '{message}'
        self.style = style
        self.FORMATS = {
            logging.DEBUG: self.GREEN + self.fmt + self.RESET,
            logging.INFO: self.INVERT + self.YELLOW + self.fmt + self.RESET,
            logging.WARNING: self.BLUE + self.plain_fmt + self.RESET,
            logging.ERROR: self.RED + self.fmt + self.RESET,
            logging.CRITICAL: self.INVERT + self.RED + self.fmt + self.RESET
        }
        if emoji:
            warning = self.FORMATS[logging.WARNING]
            self.FORMATS[logging.WARNING] = rnd_good_emoji(2) +\
                "  " + warning + "  " + rnd_good_emoji(2)

            error = self.FORMATS[logging.ERROR]
            self.FORMATS[logging.ERROR] = rnd_bad_emoji(2) +\
                "  " + error + "  " + rnd_bad_emoji(2)

            critical = self.FORMATS[logging.CRITICAL]
            self.FORMATS[logging.CRITICAL] = rnd_bad_emoji(3) +\
                "  " + critical + "  " + rnd_bad_emoji(3)

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, style=self.style)
        return formatter.format(record)


def setup_log_record_factory():
    old_factory = logging.getLogRecordFactory()

    def record_factory(*args, **kwargs):
        shorty = args[2].replace('.py','')
        file_lineno = f'{shorty}@{args[3]}'
        custom = list(args)
        custom[2] = file_lineno
        custom[3] = ''
        record = old_factory(*tuple(custom), **kwargs)
        record.custom_attribute = 0xdecafbad
        return record

    logging.setLogRecordFactory(record_factory)


def setup_console_handler(verbose, dry_run):
    if verbose:
        level = logging.DEBUG
    elif dry_run:
        level = logging.INFO
    else:
        level = logging.WARNING
    logger.__dry_run__ = False

    logger.setLevel(level)
    console = logging.StreamHandler(sys.stderr)
    console.setLevel(level)
    #format_str = '{filename:<20}{lineno}  {message}'
    format_str = '{filename:<16} {message}'
    formatter = CustomFormatter(format_str, style="{")
    console.setFormatter(formatter)
    logger.addHandler(console)

    logger.debug(f' ---===::: Welcome to ican v{__version__} :::===---')
    if dry_run:
            logger.__dry_run__ = True
            logger.info('--dry-run detected - no files will be modified')

def setup_file_handler(filename):
    format_str = '%(asctime)s | %(levelname)s | %(message)s'
    date_format = '%m-%d-%Y %H:%M:%S'
    formatter = logging.Formatter(format_str, date_format)
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

def ok_to_write():
    if not logger.__dry_run__:
        return True
    logger.info('Skipping file write @ --dry-run')
    return False


logger = logging.getLogger('ican')
#setup_log_record_factory()



# -*- coding: utf-8 -*-
"""
"""
import os
import re
import subprocess
from pathlib import Path
from types import SimpleNamespace

from .log import logger
from .log import ok_to_write
from . import exceptions


#######################################
#
#   Pipeline
#
#######################################


class PipeLine(object):

    TEMPLATE = r"{{(?P<var>.*?)}}"

    def __init__(self, label=None, steps=None):
        self.label = label
        self.steps = []
        self.compiled = re.compile(PipeLine.TEMPLATE)

        if steps is None:
            logger.error('must include at least 1 step')

        if steps:
            for k, v in steps:
                logger.debug(f'{label.upper()}.{k} - {v}')
                step = SimpleNamespace(label=k, cmd=v)
                self.steps.append(step)

    def _render(self, cmd, ctx):
        """render jinja-style templates
        {{var}} = ctx['var']
        """

        result = self.compiled.sub(
            lambda m: ctx.get(
                m.group('var'),
                'MISSING'
            ),
            cmd
        )
        return result

    def _run_cmd(self, _cmd=[]):
        cmd = _cmd.split(' ')

        result = subprocess.run(
            _cmd,
            shell=True,
            capture_output=True,
            text=True
        )

        #result.stderr
        #result.stdout
        return result

    def run(self, ctx={}):
        for step in self.steps:
            cmd = self._render(step.cmd, ctx)
            label = step.label
            logger.debug(f'rendered {cmd}')
            if ok_to_write():
                result = self._run_cmd(cmd)
                logger.debug(f'result: {result.stdout}{result.stderr}')



# -*- coding: utf-8 -*-

import re
from types import SimpleNamespace
from .log import ok_to_write
from .log import logger


__version__ = '0.2'


#######################################
#
#   Version Class
#
#######################################


class Version(object):
    """
    Semver representation
    """

    BUMPABLE = ['major', 'minor', 'patch', 'prerelease', 'build']
    DEFAULT_PRERELEASE = 'beta.0'
    DEFAULT_BUILD = 'build.0'

    last_number_re = re.compile(r"(?:[^\d]*(\d+)[^\d]*)+")
    semver_re = re.compile(
        r"""
            ^(?P<major>0|[1-9]\d*)
            \.(?P<minor>0|[1-9]\d*)
            \.(?P<patch>0|[1-9]\d*)
            (?:-(?P<prerelease>
                (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
                (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
            ))?
            (?:\+(?P<build>
                [0-9a-zA-Z-]+
                (?:\.[0-9a-zA-Z-]+)*
            ))?$
        """,
        re.VERBOSE,
    )
    pep440_re = re.compile(
        r'''
        ^([1-9][0-9]*!)?
        (0|[1-9][0-9]*)
        (\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?
        (\.post(0|[1-9][0-9]*))?
        (\.dev(0|[1-9][0-9]*))?$
        ''',
        re.VERBOSE
    )

    def __init__(
        self, 
        major=0, 
        minor=0, 
        patch=0, 
        prerelease=None, 
        build=None):

        self._major = int(major)
        self._minor = int(minor)
        self._patch = int(patch)
        self._prerelease = prerelease
        self._build = build
        self._git_metadata = None

        f = SimpleNamespace(
            major = self.major,
            minor = self.minor,
            patch = self.patch,
            prerelease = self.prerelease,
            build = self.build,
            semantic = self.semantic,
            part = None
        )
        self._frozen = f

    def __str__(self):
        return self.semantic

    def __repr__(self):
        address = hex(id(self))
        return f'<Version {self.semantic} at {address}>'


##########################################
#
#  Version Parts: 
#    major, minor, patch, pre, build
#
##########################################


    @property
    def major(self):
        return self._major

    @major.setter
    def major(self, value):
        raise AttributeError("major is readonly")

    @property
    def minor(self):
        return self._minor

    @minor.setter
    def minor(self, value):
        raise AttributeError("minor is readonly")

    @property
    def patch(self):
        return self._patch

    @patch.setter
    def patch(self, value):
        raise AttributeError("patch is readonly")

    @property
    def prerelease(self):
        return self._prerelease

    @prerelease.setter
    def prerelease(self, value):
        raise AttributeError("prerelease is readonly")

    @property
    def prerelease_int(self):
        """
        Parse an int from prerelease... 
        """

        if not self.prerelease:
            return None
        match = Version.last_number_re.search(self.prerelease)
        if match:
            return int(match.group(1))

    @prerelease_int.setter
    def prerelease_int(self, value):
        raise AttributeError("prerelease_int is readonly")

    @property
    def build(self):
        return self._build

    @build.setter
    def build(self, value):
        raise AttributeError("build is readonly")

    @property
    def build_int(self):
        """
        Parse an int from build... 
        +build.102 = 102
        """

        if not self.build:
            return 0
        match = Version.last_number_re.search(self.build)
        if match:
            return int(match.group(1))

    @build_int.setter
    def build_int(self, value):
        raise AttributeError("build_int is readonly")


##########################################
#
#  Version Styles: 
#    semantic, public, pep440, git
#
##########################################


    @property
    def semantic(self):
        """
        This is the standard semantic version format.  For most
        purposes this is the best choice.
        """
        v = f'{self._major}.{self._minor}.{self._patch}'
        if self._prerelease:
            v = f'{v}-{self._prerelease}'
        if self._build:
            v = f'{v}+{self._build}'
        return v

    @semantic.setter
    def semantic(self, value):
        raise AttributeError("semantic is readonly")

    @property
    def pep440(self):
        """
        This is a loose translation to pep440 compliant.  They 
        allow more than 3 segments of ints, so we can do 
        1.2.3.899b2 where 899 is build # and b2 is prerelease
        """
        base = self.public
        prerelease = ''
        build = ''

        if self.prerelease:
            numeric = self.prerelease_int
            if numeric is None:
                prerelease = ''
            elif 'alpha' in self.prerelease.lower():
                prerelease = f'a{numeric}'
            elif 'beta' in self.prerelease.lower():
                prerelease = f'b{numeric}'
            elif 'rc' in self.prerelease.lower():
                prerelease = f'rc{numeric}'
            else:
                prerelease = f'.dev{numeric}'

        if self.build:
            build = f'.{self.build_int}'

        v = f'{base}{prerelease}{build}'
        return v

    @pep440.setter
    def pep440(self, value):
        raise AttributeError("pep440 is readonly")
    
    @property
    def public(self):
        """
        Public Version - Something simple for pypi
        Should be PEP440, but looks like they take plain
        and simple semantic versions.
        """

        return f'{self._major}.{self._minor}.{self._patch}'

    @public.setter
    def public(self, value):
        raise AttributeError("public is readonly")

    @property
    def git(self):
        """
        This is a git formatted version, made up from metadata
        retrieved with `git describe`.
        
        format will be: M.m.p-dev.<distance>+build.<commit_sha>
        ex: 4.2.0-rc.3.dev.5+fcf2c8fd
        """
        if self._git_metadata is None:
            return 'GIT METADATA NOT AVAILABLE'

        tag = self._git_metadata.tag
        dirty = self._git_metadata.dirty
        commit_sha = self._git_metadata.commit_sha
        distance = self._git_metadata.distance

        # Add distance in prerelease. Check for existing prerelease
        if distance and self.prerelease:
            pre = f'-{self.prerelease}.dev.{distance}'
        elif distance:
            pre = f'-dev.{distance}'
        elif dirty:
            pre = f'-dev.DIRTY'
        else:
            pre = ''

        # Construct build metadata with git sha + tracked build
        if commit_sha:
            build = f'{commit_sha}.{self.build_int}'
        else:
            build = f'build.{self.build_int}'

        # Add build metadata to version
        v = f'{tag}{pre}+{build}'
        return v

    @git.setter
    def git(self, value):
        raise AttributeError("git is readonly")


##########################################
#
#  Descriptors: original, new_release
#
##########################################


    @property
    def original(self):
        """
        Use the data stored in _frozen to return a Version instance
        of the original data, in case a bump has been made and some
        type of comparison is desired.

        _frozen.part: the part of the original version that was
        bumped.  Will be None until a bump occurs.
        """
        
        return Version.parse(self._frozen.semantic)

    @original.setter
    def original(self, value):
        raise AttributeError("original is readonly")

    @property
    def new_release(self):
        if not self._frozen:
            return None
        
        part = self._frozen.part
        if part in ['major', 'minor', 'patch']:
            logger.debug('* VERSION: release pipeline will trigger')
            return True
        return False

    @new_release.setter
    def new_release(self, value):
        raise AttributeError("new_release is readonly")

    @property
    def tag(self):
        return f'v{self.public}'

    @tag.setter
    def tag(self, value):
        raise AttributeError("tag is readonly")


##########################################
#
#  Version methods
#
##########################################


    def is_canonical(self):
        return Version.pep440_re.match(self.pep440) is not None

    def bump(self, part=None):
        """
        Exposed bump method in the public api.
        Arguments:
            part: Which part of the semantic version to bump.
            Still valid if blank because the tracked build
            number will be incremented.
        """
        part = part.lower()

        if part not in Version.BUMPABLE:
            raise ValueError(f"{part} is not bumpable")

        # Record the bumped part
        self._frozen.part = part
        # Always increment the build number
        self.increment_build()

        # Find the part-specifric bump function
        bump_method = getattr(self, f'bump_{part}')
        bump_method()

    def bump_major(self):
        """
        Bump method for bumping MAJOR
        """

        self._major = self._major + 1
        self._minor = 0
        self._patch = 0
        self._prerelease = None

    def bump_minor(self):
        """
        Bump method for bumping MINOR
        """

        self._minor = self._minor + 1
        self._patch = 0
        self._prerelease = None
        
    def bump_patch(self):
        """
        Bump method for bumping PATCH
        """

        self._patch = self._patch + 1
        self._prerelease = None

    def bump_prerelease(self):
        """
        Bump method for bumping PRERELEASE
        """

        prerelease = Version.increment_string(
            self._prerelease or Version.DEFAULT_PRERELEASE
        )
        self._prerelease = prerelease

    def bump_build(self):
        """
        """
        pass

    def increment_build(self):
        """
        Always increment build number
        """

        build = Version.increment_string(
            self._build or Version.DEFAULT_BUILD
        )
        self._build = build

    @classmethod
    def increment_string(cls, string):
        """
        Look for the last sequence of number(s) in a string and increment.
        :param string: the string to search for.
        :return: the incremented string
        Source:
        http://code.activestate.com/recipes/442460-increment-numbers-in-a-string/#c1
        """

        match = cls.last_number_re.search(string)
        if match:
            next_ = str(int(match.group(1)) + 1)
            start, end = match.span(1)
            string = string[: max(end - len(next_), start)] + next_ + string[end:]
        return string

    @classmethod
    def parse(cls, version):
        """
        Parse version string to a Version instance.
        """

        match = cls.semver_re.match(version)
        if match is None:
            raise ValueError(f"{version} is not valid semantic version string")

        matched_parts = match.groupdict()
        return cls(**matched_parts)




# -*- coding: utf-8 -*-

import os
import re
from pathlib import Path

from . import logger

version_re = re.compile(
    r"__version__\s*=\s*(?P<quote>[\'\"])(?P<version>.+)(?P=quote)"
)


#######################################
#
#   Misc utils
#
#######################################

def read_file(f):
    return Path(f).read_text()

def parse_version_from_file(f):
    """
    Often times the only record of current_version is inside
    a source file such as __init__.py
    """

    file_txt = read_file(f)
    #self.debug(f'Opened and read file: {f} to memory.')

    match = version_re.search(file_txt, re.MULTILINE)
    if match is None:
        raise ValueError('Could not find __version__ in __init__.py.')
    #self.debug(f'Found potential version: {match.group(1)}.')

    return match.group(1)

def search_for_files(f, root):
    # First we look in current_dir + all subdirs
    root = config_root
    #self.debug(f'Searching for {f} in dir {root}')
    matches = [x for x in Path(root).rglob(f)]
    if len(matches) > 0:
        return matches


# -*- coding: utf-8 -*-

import os
from pathlib import Path
from configparser import ConfigParser
from configparser import DuplicateSectionError
from types import SimpleNamespace

from .source import SourceCode
from .pipeline import PipeLine
from .log import logger
from .log import ok_to_write
from .log import setup_file_handler
from .exceptions import NoConfigFound
from .exceptions import DuplicateConfigSections
from .exceptions import ConfigWriteError


#######################################
#
#   Config Class
#
#######################################


class Config(object):
    """
    Object which will orchestrate entire program
    """

    default_ver = dict(current = '0.1.0')
    default_file = dict(file = '*.py',
        style = 'semantic',
        variable = '__version__')

    CONFIG_FILE = '.ican'
    DEFAULT_CONFIG = dict(
        version=default_ver,
        file1=default_file
    )

    def __init__(self, git_root=None):
        self.config_file = None
        self.git_root = git_root
        self.ran_from = Path.cwd()
        self.parser = ConfigParser()

        self.current_version = None
        self.log_file = None
        self.source_files = []
        self.pipelines = {}
        self.aliases = {}

    @property
    def path(self):
        if self.config_file:
            return self.config_file.parent
        return None

    def ch_dir_root(self):
        """chdir to the config root, so if we run from another dir, relative
        file paths, etc still work as expected.
        """
        if self.git_root:
            os.chdir(str(self.git_root).rstrip('\n'))
        elif self.path:
            os.chdir(str(self.path).rstrip('\n'))

    def save(self):
        if ok_to_write():
            if not self.config_file:
                f = Path(self.ran_from, Config.CONFIG_FILE)
                self.config_file = f
            try:
                self.parser.write(open(self.config_file, "w"))
                logger.debug('wrote config file')
            except Exception as e:
                raise ConfigWriteError(e)
        return

    def persist_version(self, version_str):
        """
        Update the version in the config file then write it so we know the
        new version next time
        """
        logger.debug(f'persisting version - {version_str}')

        self.parser.set('version', 'current', version_str)
        self.save()
        return None

    def init(self):
        """
        Set default config and save
        """
        logger.debug(f'command init - setting default config')
        self.parser.read_dict(Config.DEFAULT_CONFIG)
        self.save()
        return

    def search_for_config(self):
        logger.debug(f'searching for config file')
        f = Config.CONFIG_FILE
        dirs = [
            self.git_root,
            self.ran_from,
            self.ran_from.parent,
            self.ran_from.parent.parent
        ]
        for d in dirs:
            if d is None:
                continue
            c = Path(d, f)
            if c.exists():
                try:
                    self.parser.read(c)
                except DuplicateSectionError:
                    raise DuplicateConfigSections()
                self.config_file = c
                logger.debug(f'config found @ {c}')
                break
        else:
            raise NoConfigFound()
        return

    def parse(self):
        """
        We search for a config in several locations.  Also, the user may
        specify default config, which we load here as well.
        """

        # By now we may have git_root or config_root
        self.ch_dir_root()

        # Current version
        self.current_version = self.parser.get(
            'version', 'current', fallback='0.1.0'
        )

        # OPTIONS - log file setup
        self.log_file = self.parser.get('options', 'log_file', fallback=None)
        if self.log_file:
            setup_file_handler(self.log_file)

        self.parse_source_files()
        self.parse_pipelines()
        self.parse_aliases()

        return

    def parse_aliases(self):
        # aliases
        if not self.parser.has_section('aliases'):
            return

        for alias, built_in in self.parser.items('aliases'):
            cmd_with_args = built_in.strip().split(' ')
            self.aliases[alias] = cmd_with_args
        return

    def parse_pipelines(self):
        # Pipelines
        for s in self.parser.sections():
            if not s.startswith('pipeline:'):
                # Not interested in this section
                continue

            label = s.split(':')[1].strip().lower()
            logger.debug(f'parsing {label.upper()} pipeline')
            list_tuples = self.parser.items(s)

            pl = PipeLine(label=label, steps=list_tuples)
            self.pipelines[label] = pl

        return

    def parse_source_files(self):
        # FILES TO WRITE
        for s in self.parser.sections():
            if not s.startswith('file:'):
                # Not interested in this section
                continue

            label = s.split(':')[1].strip().lower()
            file = self.parser.get(s, 'file', fallback=None)
            variable = self.parser.get(s, 'variable', fallback=None)
            style = self.parser.get(s, 'style', fallback='semantic')
            regex = self.parser.get(s, 'regex', fallback=None)

            # Instead of raising exp, we can just look for more files
            if file is None:
                logger.debug(f'skipping source - missing file ({label})')
                continue
            elif variable is None and regex is None:
                logger.debug(f'skipping source - missing variable/regex')
                continue

            logger.debug(f'parsing version file {label.upper()}[{file}]')

            # Case with *.py for all python files
            if '*' in file:
                files = self.search_for_files(file)
            else:
                files = [file]
            for f in files:
                u = SourceCode(
                    label,
                    f,
                    style=style,
                    variable=variable,
                    regex=regex
                )
                self.source_files.append(u)

    def search_for_files(self, f):
        """
        First we look in current_dir + all subdirs
        """

        logger.debug(f'* CONFIG: searching for - {f}')

        root = self.path
        #self.debug(f'Searching for {f} in dir {root}')
        matches = [x for x in Path(root).rglob(f)]
        if len(matches) > 0:
            logger.debug(f'found: {len(matches)} files')
            return matches
        return None
